from __future__ import annotations

import abc
import dataclasses
import typing

import ldap.dn

from ldaptool._utils.dninfo import DNInfo


class Step(abc.ABC):
    __slots__ = ()

    @abc.abstractmethod
    def step(self, value: str) -> str:
        ...


def _args_to_slice(args: str) -> slice:
    args = args.strip()
    if not args:
        return slice(None)
    params: list[typing.Optional[int]] = []
    for arg in args.split(":"):
        arg = arg.strip()
        if arg:
            params.append(int(arg))
        else:
            params.append(None)
    if len(params) == 1:
        assert isinstance(params[0], int)
        ndx = params[0]
        if ndx == -1:
            return slice(ndx, None)  # from last element to end - still exactly one element
        # this doesn't work for ndx == -1: slice(-1, 0) is always empty. otherwise it should return [ndx:][:1].
        return slice(ndx, ndx + 1)
    return slice(*params)


@dataclasses.dataclass(slots=True)
class MaxLength(Step):
    limit: int

    def step(self, value: str) -> str:
        if not self.limit or len(value) <= self.limit:
            return value
        return value[: self.limit - 1] + "…"


@dataclasses.dataclass(slots=True)
class DNDomain(Step):
    def __init__(self, args: str) -> None:
        if args:
            raise ValueError(":domain doesn't support an argument")

    def step(self, value: str) -> str:
        try:
            dninfo = DNInfo(dn=value)
        except Exception:
            # not a valid DN -> no processing
            return value
        return dninfo.domain


@dataclasses.dataclass(slots=True)
class DNPath(Step):
    path_slice: slice

    def __init__(self, args: str) -> None:
        self.path_slice = _args_to_slice(args)

    def step(self, value: str) -> str:
        try:
            dninfo = DNInfo(dn=value)
        except Exception:
            # not a valid DN -> no processing
            return value
        return dninfo.sliced_path(self.path_slice)


@dataclasses.dataclass(slots=True)
class DNFullPath(Step):
    path_slice: slice

    def __init__(self, args: str) -> None:
        self.path_slice = _args_to_slice(args)

    def step(self, value: str) -> str:
        try:
            dninfo = DNInfo(dn=value)
        except Exception:
            # not a valid DN -> no processing
            return value
        return dninfo.sliced_full_path(self.path_slice)


@dataclasses.dataclass(slots=True)
class DNSlice(Step):
    slice: slice

    def __init__(self, args: str) -> None:
        self.slice = _args_to_slice(args)

    def step(self, value: str) -> str:
        try:
            dninfo = DNInfo(dn=value)
        except Exception:
            # not a valid DN -> no processing
            return value
        return ldap.dn.dn2str(dninfo.parts[self.slice])  # type: ignore


_STEPS: dict[str, typing.Callable[[str], Step]] = {
    "domain": DNDomain,
    "path": DNPath,
    "fullpath": DNFullPath,
    "dnslice": DNSlice,
}


class InvalidStep(Exception):
    pass


@dataclasses.dataclass(slots=True)
class PostProcess:
    steps: list[Step]

    def process(self, value: str) -> str:
        for step in self.steps:
            value = step.step(value)
        return value


def parse_steps(steps: str) -> PostProcess:
    result: list[Step] = []

    cur_id_start = 0
    cur_args_start = -1
    current_id = ""
    current_args = ""
    count_brackets = 0
    step_done = False

    def handle_step() -> None:
        nonlocal cur_id_start, cur_args_start, current_id, current_args, step_done
        assert step_done

        step_i = _STEPS.get(current_id, None)
        if step_i is None:
            try:
                max_len = int(current_id)
                result.append(MaxLength(max_len))
            except ValueError:
                raise InvalidStep(f"Unknown post-processing step {current_id!r}")
        else:
            result.append(step_i(current_args))

        cur_id_start = pos + 1
        cur_args_start = -1
        current_id = ""
        current_args = ""
        step_done = False

    for pos, char in enumerate(steps):
        if step_done:
            if char != ":":
                raise InvalidStep(f"Require : after step, found {char!r} at pos {pos}")
            handle_step()
        elif char == "[":
            if count_brackets == 0:
                # end of identifier
                current_id = steps[cur_id_start:pos]
                cur_args_start = pos + 1
            count_brackets += 1
        elif char == "]":
            count_brackets -= 1
            if count_brackets == 0:
                current_args = steps[cur_args_start:pos]
                step_done = True
        elif count_brackets:
            continue
        elif not char.isalnum():
            raise InvalidStep(f"Expecting either alphanumeric, ':' or '[', got {char!r} at {pos}")

    if not step_done:
        current_id = steps[cur_id_start:]
        if current_id:
            step_done = True

    if step_done:
        handle_step()

    return PostProcess(result)
